import dataclasses
import enum
import textwrap
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Annotated,
    Any,
    Awaitable,
    Callable,
    Final,
    Generic,
    Iterable,
    Mapping,
    TypeAlias,
    TypeVar,
)
from uuid import UUID

import aiohttp_cors
from aiohttp import web
from aiohttp.typedefs import Middleware
from pydantic import AnyUrl, BaseModel, Field

# FIXME: merge majority of common definitions to ai.backend.common when ready


class FrontendMode(str, enum.Enum):
    PORT = "port"


class ProxyProtocol(str, enum.Enum):
    HTTP = "http"
    GRPC = "grpc"
    HTTP2 = "h2"
    TCP = "tcp"
    PREOPEN = "preopen"


class AppMode(str, enum.Enum):
    INTERACTIVE = "interactive"
    INFERENCE = "inference"


@dataclass
class Slot:
    frontend_mode: FrontendMode
    in_use: bool
    port: int | None
    circuit_id: UUID | None


class EventLoopType(str, enum.Enum):
    UVLOOP = "uvloop"
    ASYNCIO = "asyncio"


class DigestModType(str, enum.Enum):
    SHA1 = "sha1"
    SHA224 = "sha224"
    SHA256 = "sha256"
    SHA384 = "sha384"
    SHA512 = "sha512"


WebRequestHandler: TypeAlias = Callable[
    [web.Request],
    Awaitable[web.StreamResponse],
]
WebMiddleware: TypeAlias = Middleware

CORSOptions: TypeAlias = Mapping[str, aiohttp_cors.ResourceOptions]
AppCreator: TypeAlias = Callable[
    [CORSOptions],
    tuple[web.Application, Iterable[WebMiddleware]],
]


class RouteInfo(BaseModel):
    session_id: UUID
    session_name: Annotated[str | None, Field(default=None)]
    kernel_host: str
    kernel_port: int
    protocol: ProxyProtocol
    traffic_ratio: Annotated[float, Field(default=1.0)]


@dataclass
class PortFrontendInfo:
    port: int


@dataclass
class InteractiveAppInfo:
    user_id: UUID


@dataclass
class InferenceAppInfo:
    endpoint_id: UUID


class Circuit(BaseModel):
    id: Annotated[UUID, Field(UUID, description="ID of circuit.")]

    app: Annotated[
        str,
        Field(
            str,
            description="Name of the Backend.AI Kernel app circuit is hosting. Can be a blank string if circuit is referencing an inference app.",
        ),
    ]
    protocol: Annotated[
        ProxyProtocol, Field(ProxyProtocol, description="Protocol of the Backend.AI Kernel app.")
    ]
    worker: Annotated[UUID, Field(UUID, description="ID of the worker hosting the circuit.")]

    app_mode: Annotated[AppMode, Field(AppMode, description="Application operation mode.")]
    frontend_mode: Annotated[
        FrontendMode, Field(FrontendMode, description="Frontend type of worker.")
    ]

    envs: dict[str, Any]
    arguments: str | None

    open_to_public: Annotated[
        bool,
        Field(
            bool,
            description=textwrap.dedent(
                """
                Shows if the circuit is open to public.
                For interactive apps, this set as true means users without authorization cookie set will also be able to access application.
                For inference apps it means that API will work without authorization token passed.
                """
            ),
        ),
    ]

    allowed_client_ips: Annotated[
        str | None,
        Field(
            str | None,
            description="Comma separated list of CIDRs accepted as traffic source. null means the circuit is accessible anywhere.",
        ),
    ]

    port: Annotated[
        int,
        Field(int, description="Occupied worker port."),
    ]

    user_id: Annotated[UUID | None, Field(UUID | None, description="Session owner's UUID.")]
    access_key: Annotated[str | None, Field(str | None, description="Session owner's access key.")]
    endpoint_id: Annotated[
        UUID | None,
        Field(
            UUID | None, description="Model service's UUID. Only set if `app_mode` is inference."
        ),
    ]

    route_info: Annotated[
        list[RouteInfo], Field(list[RouteInfo], description="List of kernel access information.")
    ]
    session_ids: list[UUID]

    created_at: datetime
    updated_at: datetime

    @property
    def app_info(self) -> InteractiveAppInfo | InferenceAppInfo:
        match self.app_mode:
            case AppMode.INTERACTIVE:
                assert self.user_id
                return InteractiveAppInfo(self.user_id)
            case AppMode.INFERENCE:
                assert self.endpoint_id
                return InferenceAppInfo(self.endpoint_id)
            case _:
                raise KeyError(f"{self.app_mode} is not a valid app mode")


class Token(BaseModel):
    login_session_token: str | None
    kernel_host: str
    kernel_port: int
    session_id: UUID
    user_uuid: UUID
    group_id: UUID
    access_key: str
    domain_name: str


class SessionConfig(BaseModel):
    id: Annotated[UUID | None, Field(default=None)]
    user_uuid: UUID
    group_id: UUID
    access_key: Annotated[str | None, Field(default=None)]
    domain_name: str


class EndpointConfig(BaseModel):
    id: UUID
    existing_url: AnyUrl | None


TBaseModel = TypeVar("TBaseModel", bound=BaseModel)


@dataclass
class PydanticResponse(Generic[TBaseModel]):
    response: TBaseModel
    headers: dict[str, Any] = dataclasses.field(default_factory=lambda: {})
    status: int = 200


PERMIT_COOKIE_NAME: Final[str] = "appproxy_permit"

TCircuitKey = TypeVar("TCircuitKey", int, str)



from typing import Iterable
from uuid import UUID

import aiohttp_cors
from aiohttp import web

from ai.backend.wsproxy.types import (
    CORSOptions,
    PydanticResponse,
    WebMiddleware,
)

from ..defs import RootContext
from ..exceptions import ObjectNotFound
from .types import StubResponseModel
from .utils import auth_required, pydantic_api_response_handler


@auth_required("worker")
@pydantic_api_response_handler
async def delete_circuit(request: web.Request) -> PydanticResponse[StubResponseModel]:
    """
    Removes circuit record from wsproxy.
    """
    root_ctx: RootContext = request.app["_root.context"]
    circuit_id = UUID(request.match_info["circuit_id"])

    try:
        circuit = root_ctx.proxy_frontend.circuits[circuit_id]
    except KeyError:
        raise ObjectNotFound(object_name="Circuit")
    await root_ctx.proxy_frontend.break_circuit(circuit)

    return PydanticResponse(StubResponseModel(success=True))


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "api/circuit"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    cors.add(add_route("DELETE", "/{circuit_id}", delete_circuit))
    return app, []



import textwrap
from datetime import datetime
from typing import Annotated, Iterable
from uuid import UUID

import aiohttp_cors
import jwt
from aiohttp import web
from pydantic import AnyUrl, BaseModel, Field

from ai.backend.wsproxy.exceptions import ObjectNotFound
from ai.backend.wsproxy.types import (
    AppMode,
    CORSOptions,
    EndpointConfig,
    ProxyProtocol,
    PydanticResponse,
    RouteInfo,
    WebMiddleware,
)

from ..defs import RootContext
from ..registry import add_circuit
from ..types import SessionConfig
from .types import StubResponseModel
from .utils import (
    auth_required,
    pydantic_api_handler,
    pydantic_api_response_handler,
)


class EndpointTagConfig(BaseModel):
    session: SessionConfig
    endpoint: EndpointConfig


class InferenceAppConfig(BaseModel):
    session_id: UUID
    kernel_host: str
    kernel_port: int
    protocol: Annotated[ProxyProtocol, Field(default=ProxyProtocol.HTTP)]
    traffic_ratio: Annotated[float, Field(ge=0.0, le=1.0, default=1.0)]


class EndpointCreationRequestModel(BaseModel):
    service_name: Annotated[str, Field(description="Name of the model service.")]
    tags: Annotated[
        EndpointTagConfig,
        Field(
            description="Metadata of target model service and dependent sessions.",
        ),
    ]
    apps: Annotated[
        dict[str, list[InferenceAppConfig]],
        Field(
            description=textwrap.dedent(
                """
                key-value pair of available applications exposed by requested endpoint.
                Key should be name of the app, and value as list of host-port pairs app is bound to.
                """
            ),
        ),
    ]
    open_to_public: Annotated[
        bool,
        Field(
            default=False,
            description=textwrap.dedent(
                """
                If set to true, AppProxy will require an API token (which can be obtained from `generate_endpoint_api_token` request)
                fullfilled at request header.
                """
            ),
        ),
    ]

    port: Annotated[int | None, Field(default=None, description="Preferred port number.")]
    subdomain: Annotated[str | None, Field(default=None, description="Preferred subdomain name.")]


class EndpointCreationResponseModel(BaseModel):
    endpoint: AnyUrl


@auth_required("manager")
@pydantic_api_handler(EndpointCreationRequestModel)
async def create_or_update_endpoint(
    request: web.Request, params: EndpointCreationRequestModel
) -> PydanticResponse[EndpointCreationResponseModel]:
    """
    Creates or updates an inference circuit.
    """
    root_ctx: RootContext = request.app["_root.context"]
    endpoint_id = UUID(request.match_info["endpoint_id"])

    app_names = list(params.apps.keys())
    if len(app_names) > 0:
        app = list(params.apps.keys())[0]
        routes = [RouteInfo(**r.model_dump()) for r in params.apps[app]]
    else:
        app = ""
        routes = []

    try:
        circuit = root_ctx.proxy_frontend.get_circuit_by_endpoint_id(endpoint_id)
        circuit.route_info = routes
        circuit.session_ids = [r.session_id for r in routes]
        circuit.open_to_public = params.open_to_public
        await root_ctx.proxy_frontend.update_circuit_route_info(circuit, routes)
    except ObjectNotFound:
        circuit = await add_circuit(
            root_ctx,
            params.tags.session,
            params.tags.endpoint,
            app,
            params.apps[app][0].protocol if app else ProxyProtocol.HTTP,
            AppMode.INFERENCE,
            routes,
            open_to_public=params.open_to_public,
        )
    endpoint_url = f"http://{root_ctx.local_config.wsproxy.advertised_host}:{circuit.port}"

    return PydanticResponse(EndpointCreationResponseModel(endpoint=AnyUrl(endpoint_url)))


@auth_required("manager")
@pydantic_api_response_handler
async def remove_endpoint(request: web.Request) -> PydanticResponse[StubResponseModel]:
    """
    Deassociates inference circuit from system.
    """
    root_ctx: RootContext = request.app["_root.context"]

    endpoint_id = UUID(request.match_info["endpoint_id"])

    circuit = root_ctx.proxy_frontend.get_circuit_by_endpoint_id(endpoint_id)
    await root_ctx.proxy_frontend.break_circuit(circuit)

    return PydanticResponse(StubResponseModel(success=True))


class EndpointAPITokenGenerationRequestModel(BaseModel):
    user_uuid: UUID
    """
    Token requester's user UUID.
    """
    exp: datetime
    """
    Expiration date of token.
    """


class EndpointAPITokenResponseModel(BaseModel):
    token: str


@auth_required("manager")
@pydantic_api_handler(EndpointAPITokenGenerationRequestModel)
async def generate_endpoint_api_token(
    request: web.Request, params: EndpointAPITokenGenerationRequestModel
) -> PydanticResponse[EndpointAPITokenResponseModel]:
    """
    Creates and returns API token required for execution of model service apps hosted by AppProxy.
     This API is meant to be called from Backend.AI manager rather than model service callee itself.
    """
    root_ctx: RootContext = request.app["_root.context"]

    endpoint_id = UUID(request.match_info["endpoint_id"])

    circuit = root_ctx.proxy_frontend.get_circuit_by_endpoint_id(endpoint_id)
    payload = circuit.model_dump(mode="json")
    payload["config"] = {}
    payload["app_url"] = f"http://{root_ctx.local_config.wsproxy.advertised_host}:{circuit.port}"
    payload["user"] = str(params.user_uuid)
    payload["exp"] = params.exp
    encoded_jwt = jwt.encode(
        payload, root_ctx.local_config.wsproxy.jwt_encrypt_key, algorithm="HS256"
    )
    return PydanticResponse(EndpointAPITokenResponseModel(token=encoded_jwt))


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "v2/endpoints"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    add_route = app.router.add_route
    cors.add(app.router.add_resource(r""))
    cors.add(add_route("POST", "/{endpoint_id}", create_or_update_endpoint))
    cors.add(add_route("DELETE", "/{endpoint_id}", remove_endpoint))
    cors.add(add_route("POST", "/{endpoint_id}/token", generate_endpoint_api_token))
    return app, []



import urllib.parse
from logging import LoggerAdapter
from typing import Iterable
from uuid import UUID

import jwt
from aiohttp import web
from pydantic import AnyUrl, BaseModel

from ..config import ServerConfig
from ..defs import RootContext
from ..exceptions import (
    InvalidAPIParameters,
)
from ..types import (
    PERMIT_COOKIE_NAME,
    Circuit,
    CORSOptions,
    InteractiveAppInfo,
    ProxyProtocol,
    PydanticResponse,
    WebMiddleware,
)
from ..utils import calculate_permit_hash
from .utils import pydantic_api_handler


def generate_proxy_url(local_config: ServerConfig, protocol: str, circuit: Circuit) -> str:
    config = local_config.wsproxy
    if config.advertised_proxy_port_range:
        idx = config.bind_proxy_port_range.index(circuit.port)
        port = config.advertised_proxy_port_range[idx]
    else:
        port = circuit.port
    return f"{protocol}://{config.advertised_host}:{port}"


class ProxySetupRequestModel(BaseModel):
    token: str


class ProxySetupResponseModel(BaseModel):
    redirect: AnyUrl
    redirectURI: AnyUrl


@pydantic_api_handler(ProxySetupRequestModel)
async def setup(
    request: web.Request, params: ProxySetupRequestModel
) -> web.StreamResponse | PydanticResponse[ProxySetupResponseModel]:
    log: LoggerAdapter = request["log"]

    try:
        root_ctx: RootContext = request.app["_root.context"]
        jwt_body = jwt.decode(
            params.token, root_ctx.local_config.wsproxy.jwt_encrypt_key, algorithms=["HS256"]
        )
        requested_circuit_id = UUID(jwt_body["circuit"])

        circuit = root_ctx.proxy_frontend.get_circuit_by_id(requested_circuit_id)

        if not isinstance(circuit.app_info, InteractiveAppInfo):
            raise InvalidAPIParameters("E20011: Not supported for inference apps")

        # Web browsers block redirect between cross-origins if Access-Control-Allow-Origin value is set to a concrete Origin instead of wildcard;
        # Hence we need to send "*" as allowed origin manually, instead of benefiting from aiohttp-cors
        cors_headers = {
            "Access-Control-Allow-Origin": "*",
            "Access-Control-Allow-Headers": "*",
            "Access-Control-Expose-Headers": "*",
        }
        match circuit.protocol:
            case ProxyProtocol.HTTP:
                protocol = "http"
                response = web.HTTPPermanentRedirect(
                    generate_proxy_url(root_ctx.local_config, protocol, circuit),
                    headers=cors_headers,
                )
                response.set_cookie(
                    PERMIT_COOKIE_NAME,
                    calculate_permit_hash(
                        root_ctx.local_config.wsproxy.permit_hash_key, circuit.app_info.user_id
                    ),
                )
                return response
            case ProxyProtocol.TCP:
                protocol = "tcp"
                queryparams = {
                    "directTCP": "true",
                    "auth": params.token,
                    "proto": protocol,
                    "gateway": generate_proxy_url(root_ctx.local_config, protocol, circuit),
                }
                if jwt_body["redirect"]:
                    return web.HTTPPermanentRedirect(
                        f"http://localhost:45678/start?{urllib.parse.urlencode(queryparams)}",
                        headers=cors_headers,
                    )
                else:
                    return PydanticResponse(
                        ProxySetupResponseModel(
                            redirect=AnyUrl(
                                f"http://localhost:45678/start?{urllib.parse.urlencode(queryparams)}"
                            ),
                            redirectURI=AnyUrl(
                                f"http://localhost:45678/start?{urllib.parse.urlencode(queryparams)}"
                            ),
                        ),
                        headers=cors_headers,
                    )
            case _:
                raise InvalidAPIParameters("E20002: Protocol not available as interactive app")
    except:
        log.exception("")
        raise


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "setup"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    add_route = app.router.add_route
    add_route("GET", "", setup)
    return app, []



import logging
from abc import ABCMeta
from dataclasses import dataclass
from typing import Any

import aiohttp
from yarl import URL

from ai.backend.logging import BraceStyleAdapter
from ai.backend.wsproxy.defs import RootContext
from ai.backend.wsproxy.types import Circuit

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@dataclass
class HttpRequest:
    method: str
    path: str | URL
    headers: dict[str, Any]
    body: aiohttp.StreamReader


class AbstractBackend(metaclass=ABCMeta):
    root_context: RootContext
    circuit: Circuit

    def __init__(self, root_context: RootContext, circuit: Circuit) -> None:
        self.root_context = root_context
        self.circuit = circuit



from pathlib import Path
from typing import Self

import click

from ai.backend.logging import AbstractLogger, LocalLogger, LogLevel

from ..config import ServerConfig
from ..config import load as load_config


class CLIContext:
    _local_config: ServerConfig | None
    _logger: AbstractLogger

    def __init__(self, config_path: Path, log_level: LogLevel) -> None:
        self.config_path = config_path
        self.log_level = log_level
        self._local_config = None

    @property
    def local_config(self) -> ServerConfig:
        # Lazy-load the configuration only when requested.
        if self._local_config is None:
            self._local_config = load_config(self.config_path, self.log_level)
        return self._local_config

    def __enter__(self) -> Self:
        # The "start-server" command is injected by ai.backend.cli from the entrypoint
        # and it has its own multi-process-aware logging initialization.
        # If we duplicate the local logging with it, the process termination may hang.
        click_ctx = click.get_current_context()
        if click_ctx.invoked_subcommand != "start-server":
            self._logger = LocalLogger({
                "level": self.log_level,
                "pkg-ns": {
                    "": LogLevel.WARNING,
                    "ai.backend": self.log_level,
                },
            })
            self._logger.__enter__()
        return self

    def __exit__(self, *exc_info) -> None:
        click_ctx = click.get_current_context()
        if click_ctx.invoked_subcommand != "start-server":
            self._logger.__exit__()




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/ai/backend/wsproxy:dist

from setuptools import setup

setup(**{
    'author': 'Lablup Inc. and contributors',
    'classifiers': [
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.13',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
    ],
    'description': 'Backend.AI WSProxy',
    'entry_points': {
        'backendai_cli_v10': [
            'wsproxy = ai.backend.wsproxy.cli.__main__:main',
            'wsproxy.start-server = ai.backend.wsproxy.server:main',
        ],
    },
    'install_requires': (
        'Jinja2~=3.1.6',
        'PyJWT~=2.10.1',
        'PyYAML~=6.0',
        'aiohttp_cors~=0.8.1',
        'aiohttp_jinja2~=1.6',
        'aiohttp~=3.11.16',
        'aiomonitor~=0.7.0',
        'aiotools~=1.9.0',
        'attrs>=25.3',
        'backend.ai-common==25.6.12rc1',
        'backend.ai-logging==25.6.12rc1',
        'backend.ai-plugin==25.6.12rc1',
        'click~=8.1.7',
        'pydantic[email]~=2.11.3',
        'pyhumps~=3.8.0',
        'setproctitle~=1.3.5',
        'tomlkit~=0.13.2',
        'types-Jinja2',
        'types-PyYAML',
        'uvloop~=0.21; sys_platform != "Windows"',
        'yarl~=1.19.0',
    ),
    'license': 'LGPLv3',
    'long_description': """# Backend.AI WSProxy

""",
    'long_description_content_type': 'text/markdown',
    'name': 'backend.ai-wsproxy',
    'namespace_packages': (
    ),
    'package_data': {
        'ai.backend.wsproxy': (
            'VERSION',
        ),
        'ai.backend.wsproxy.templates': (
            '__init__.py',
            'error.jinja2',
        ),
    },
    'packages': (
        'ai.backend.wsproxy',
        'ai.backend.wsproxy.api',
        'ai.backend.wsproxy.cli',
        'ai.backend.wsproxy.proxy.backend',
        'ai.backend.wsproxy.proxy.frontend',
        'ai.backend.wsproxy.proxy.frontend.http',
        'ai.backend.wsproxy.templates',
    ),
    'project_urls': {
        'Documentation': 'https://docs.backend.ai/',
        'Source': 'https://github.com/lablup/backend.ai',
    },
    'python_requires': '>=3.13,<3.14',
    'url': 'https://github.com/lablup/backend.ai',
    'version': '25.6.12rc1',
    'zip_safe': False,
})



import urllib.parse
from logging import LoggerAdapter
from typing import Annotated, Iterable
from uuid import UUID

import aiohttp_cors
import jwt
from aiohttp import web
from pydantic import AnyUrl, BaseModel, Field

from ..defs import RootContext
from ..exceptions import (
    GenericForbidden,
    InvalidCredentials,
    ObjectNotFound,
)
from ..registry import add_circuit
from ..types import (
    AppMode,
    Circuit,
    CORSOptions,
    ProxyProtocol,
    PydanticResponse,
    RouteInfo,
    SessionConfig,
    WebMiddleware,
)
from ..utils import is_permit_valid, mime_match
from .types import ConfRequestModel, StubResponseModel
from .utils import pydantic_api_handler, pydantic_api_response_handler


class AddRequestModel(BaseModel):
    app: str
    protocol: ProxyProtocol
    envs: Annotated[dict[str, str | int | None], Field(default={})]
    args: Annotated[str | None, Field(default=None)]
    open_to_public: Annotated[bool, Field(default=False)]
    allowed_client_ips: Annotated[str | None, Field(default=None)]
    redirect: Annotated[str, Field(default="")]
    no_reuse: Annotated[bool, Field(default=False)]

    port: Annotated[int | None, Field(default=None)]
    subdomain: Annotated[str | None, Field(default=None)]


class ProxyRequestModel(AddRequestModel):
    token: str
    session_id: UUID


class AddResponseModel(BaseModel):
    code: int
    url: AnyUrl


class ProxyResponseModel(BaseModel):
    redirect_url: AnyUrl
    reuse: bool


class TokenBodyModel(ConfRequestModel):
    exp: int


@pydantic_api_response_handler
async def check_session_existence(request: web.Request) -> PydanticResponse[StubResponseModel]:
    root_ctx: RootContext = request.app["_root.context"]
    try:
        session_id = UUID(request.match_info["session_id"])
    except ValueError:
        raise ObjectNotFound(object_name="Circuit")

    token = request.match_info["token"]

    for _, circuit in root_ctx.proxy_frontend.circuits.items():
        if session_id in circuit.session_ids and (
            token == circuit.access_key or token == circuit.user_id
        ):
            break
    else:
        raise ObjectNotFound(object_name="Circuit")

    return PydanticResponse(StubResponseModel(success=True))


@pydantic_api_response_handler
async def delete_circuit_by_session(request: web.Request) -> PydanticResponse[StubResponseModel]:
    root_ctx: RootContext = request.app["_root.context"]
    try:
        session_id = UUID(request.match_info["session_id"])
    except ValueError:
        raise ObjectNotFound(object_name="Circuit")

    token = request.match_info["token"]

    for _, circuit in root_ctx.proxy_frontend.circuits.items():
        if session_id in circuit.session_ids and (
            token == circuit.access_key or token == str(circuit.user_id)
        ):
            break
    else:
        raise ObjectNotFound(object_name="Circuit")
    permit_key = request.query.get("permit_key")
    if (
        not permit_key
        or not circuit.user_id
        or (
            not is_permit_valid(
                root_ctx.local_config.wsproxy.permit_hash_key, circuit.user_id, permit_key
            )
        )
    ):
        raise GenericForbidden

    await root_ctx.proxy_frontend.break_circuit(circuit)
    return PydanticResponse(StubResponseModel(success=True))


@pydantic_api_handler(AddRequestModel, is_deprecated=True)
async def add(request: web.Request, params: AddRequestModel) -> PydanticResponse[AddResponseModel]:
    """
    Deprecated: only for legacy applications. Just call `proxy` API directly.
    Returns URL to AppProxy's `proxy` API handler.
    """
    root_ctx: RootContext = request.app["_root.context"]

    config = root_ctx.local_config.wsproxy
    base_url = (
        f"http://{config.advertised_host}:{config.advertised_api_port or config.bind_api_port}"
    )
    qdict = {
        **params.model_dump(mode="json", exclude_defaults=True),
        "token": request.match_info["token"],
        "session_id": request.match_info["session_id"],
    }
    return PydanticResponse(
        AddResponseModel(
            code=200, url=AnyUrl(f"{base_url}/v2/proxy/auth?{urllib.parse.urlencode(qdict)}")
        ),
    )


@pydantic_api_handler(ProxyRequestModel)
async def proxy(
    request: web.Request, params: ProxyRequestModel
) -> PydanticResponse[ProxyResponseModel] | web.HTTPPermanentRedirect:
    """
    Assigns worker to host proxy app and starts proxy process.
    When `Accept` HTTP header is set to `application/json` access information to worker will be handed out inside response body;
    otherwise wsproxy will try to automatically redirect callee via `Location: ` response header.
    """
    log: LoggerAdapter = request["log"]

    existing_circuit: Circuit | None = None
    circuit: Circuit
    reuse = False

    root_ctx: RootContext = request.app["_root.context"]
    token_str = params.token
    session_id = params.session_id

    token = jwt.decode(
        token_str, root_ctx.local_config.wsproxy.jwt_encrypt_key, algorithms=["HS256"]
    )

    if token["session_id"] != str(session_id):
        log.warning(
            "User requested to create app of session {} but token authorizes session {}",
            session_id,
            token["session_id"],
        )
        raise InvalidCredentials

    for _, circuit in root_ctx.proxy_frontend.circuits.items():
        if (
            token["session_id"] in circuit.session_ids
            and token["app"] == circuit.app
            and token.get("open_to_public", False) == circuit.open_to_public
            and token["allowed_client_ip"] == circuit.allowed_client_ips
        ) and not params.no_reuse:
            existing_circuit = circuit
            break

    if existing_circuit:
        reuse = True
        routes = existing_circuit.route_info
        circuit = existing_circuit
    else:
        routes = [
            RouteInfo(
                session_id=token["session_id"],
                session_name=None,
                kernel_host=token["kernel_host"],
                kernel_port=token["kernel_port"],
                protocol=params.protocol,
                traffic_ratio=1.0,
            )
        ]

        log.debug("protocol: {} ({})", params.protocol, type(params.protocol))
        if params.protocol == ProxyProtocol.PREOPEN:
            log.debug("overriding PREOPEN to HTTP")
            params.protocol = ProxyProtocol.HTTP

        circuit = await add_circuit(
            root_ctx,
            SessionConfig(
                id=token["session_id"],
                user_uuid=token["user_uuid"],
                group_id=token["group_id"],
                access_key=token["access_key"],
                domain_name=token["domain_name"],
            ),
            None,
            params.app,
            params.protocol,
            AppMode.INTERACTIVE,
            routes,
            envs=params.envs,
            args=params.args,
            open_to_public=params.open_to_public,
            allowed_client_ips=params.allowed_client_ips,
        )
        log.debug("created new circuit")

    token_to_generate_body = {
        "version": "v2",  # TODO: add support for v1
        "redirect": params.redirect,
        "circuit": str(circuit.id),
    }
    qdict = {
        "token": jwt.encode(token_to_generate_body, root_ctx.local_config.wsproxy.jwt_encrypt_key),
    }

    port = (
        root_ctx.local_config.wsproxy.advertised_api_port
        or root_ctx.local_config.wsproxy.bind_api_port
    )
    app_url = f"http://{root_ctx.local_config.wsproxy.advertised_host}:{port}/setup?{urllib.parse.urlencode(qdict)}"
    log.debug("Redirect URL created: {}", app_url)

    if mime_match(request.headers.get("accept", "text/html"), "application/json", strict=True):
        # Web browsers block redirect between cross-origins if Access-Control-Allow-Origin value is set to a concrete Origin instead of wildcard;
        # Hence we need to send "*" as allowed origin manually, instead of benefiting from aiohttp-cors
        return PydanticResponse(
            ProxyResponseModel(
                redirect_url=AnyUrl(app_url),
                # Current version of WebUI always expects this to be False for TCP protocols
                reuse=reuse if params.protocol != ProxyProtocol.TCP else False,
            ),
        )
    else:
        return web.HTTPPermanentRedirect(app_url)


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["prefix"] = "v2/proxy"
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("GET", "/{token}/{session_id}", check_session_existence))
    cors.add(app.router.add_route("GET", "/{token}/{session_id}/add", add))
    cors.add(app.router.add_route("GET", "/{token}/{session_id}/delete", delete_circuit_by_session))
    cors.add(app.router.add_route("GET", "/auth", proxy))
    return app, []
